import {Injectable} from '@angular/core';
import {BehaviorSubject, Observable, Subject} from 'rxjs';
import {HttpClient, HttpHeaders, HttpParams} from '@angular/common/http';
import {environment} from '@src/environments/environment';
import {first} from 'rxjs/operators';
import {ArgPair, Envelope, Phone, Syncable, User} from 'cic-client-meta';
import {AccountDetails} from '@app/_models';
import {LoggingService} from '@app/_services/logging.service';
import {TokenService} from '@app/_services/token.service';
import {AccountIndex} from '@app/_eth';
import {MutableKeyStore, PGPSigner, Signer} from '@app/_pgp';
import {RegistryService} from '@app/_services/registry.service';
import {CICRegistry} from 'cic-client';
import {AuthService} from './auth.service';
import {personValidation} from '@app/_helpers';
const vCard = require('vcard-parser');

@Injectable({
  providedIn: 'root'
})
export class UserService {
  headers: HttpHeaders = new HttpHeaders({'x-cic-automerge': 'client'});
  keystore: MutableKeyStore;
  signer: Signer;
  registry: CICRegistry;

  accountsMeta = [];
  accounts: any = [];
  private accountsList = new BehaviorSubject<any>(this.accounts);
  accountsSubject = this.accountsList.asObservable();

  actions: any = '';
  private actionsList = new BehaviorSubject<any>(this.actions);
  actionsSubject = this.actionsList.asObservable();

  staff: any = '';
  private staffList = new BehaviorSubject<any>(this.staff);
  staffSubject = this.staffList.asObservable();

  constructor(
    private httpClient: HttpClient,
    private loggingService: LoggingService,
    private tokenService: TokenService,
    private registryService: RegistryService,
    private authService: AuthService,
  ) {
    this.authService.init().then(() => {
      this.keystore = authService.mutableKeyStore;
      this.signer = new PGPSigner(this.keystore);
    });
    this.registry = registryService.getRegistry();
    this.registry.load();
  }

  resetPin(phone: string): Observable<any> {
    const params = new HttpParams().set('phoneNumber', phone);
    return this.httpClient.get(`${environment.cicUssdUrl}/pin`, {params});
  }

  getAccountStatus(phone: string): any {
    const params = new HttpParams().set('phoneNumber', phone);
    return this.httpClient.get(`${environment.cicUssdUrl}/pin`, {params});
  }

  getLockedAccounts(offset: number, limit: number): any {
    return this.httpClient.get(`${environment.cicUssdUrl}/accounts/locked/${offset}/${limit}`);
  }

  async changeAccountInfo(address: string, name: string, phoneNumber: string, age: string, type: string, bio: string, gender: string,
                          businessCategory: string, userLocation: string, location: string, locationType: string,
  ): Promise<any> {
    let accountInfo: any = {
      vcard: {
        fn: [{}],
        n: [{}],
        tel: [{}]
      },
      location: {}
    };
    accountInfo.vcard.fn[0].value = name;
    accountInfo.vcard.n[0].value = name.split(' ');
    accountInfo.vcard.tel[0].value = phoneNumber;
    accountInfo.products = [bio];
    accountInfo.gender = gender;
    accountInfo.age = age;
    accountInfo.type = type;
    accountInfo.category = businessCategory;
    accountInfo.location.area = location;
    accountInfo.location.area_name = userLocation;
    accountInfo.location.area_type = locationType;
    accountInfo.vcard = btoa(vCard.generate(accountInfo.vcard));
    await personValidation(accountInfo);
    const accountKey = await User.toKey(address);
    this.getAccountDetailsFromMeta(accountKey).pipe(first()).subscribe(async res => {
      const syncableAccount: Syncable = Envelope.fromJSON(JSON.stringify(res)).unwrap();
      let update = [];
      for (const prop in accountInfo) {
        update.push(new ArgPair(prop, accountInfo[prop]));
      }
      syncableAccount.update(update, 'client-branch');
      await this.updateMeta(syncableAccount, accountKey, this.headers);
    }, async error => {
      this.loggingService.sendErrorLevelMessage('Can\'t find account info in meta service', this, {error});
      const syncableAccount: Syncable = new Syncable(accountKey, accountInfo);
      await this.updateMeta(syncableAccount, accountKey, this.headers);
    });
    return accountKey;
  }

  async updateMeta(syncableAccount: Syncable, accountKey: string, headers: HttpHeaders): Promise<any> {
    const envelope = await this.wrap(syncableAccount , this.signer);
    const reqBody = envelope.toJSON();
    this.httpClient.put(`${environment.cicMetaUrl}/${accountKey}`, reqBody , { headers }).pipe(first()).subscribe(res => {
      this.loggingService.sendInfoLevelMessage(`Response: ${res}`);
    });
  }

  getAccounts(): void {
    this.httpClient.get(`${environment.cicCacheUrl}/accounts`).pipe(first()).subscribe(res => this.accountsList.next(res));
  }

  getAccountById(id: number): Observable<any> {
    return this.httpClient.get(`${environment.cicCacheUrl}/accounts/${id}`);
  }

  getActions(): void {
    this.httpClient.get(`${environment.cicCacheUrl}/actions`).pipe(first()).subscribe(res => this.actionsList.next(res));
  }

  getActionById(id: string): any {
    return this.httpClient.get(`${environment.cicCacheUrl}/actions/${id}`);
  }

  approveAction(id: string): Observable<any> {
    return this.httpClient.post(`${environment.cicCacheUrl}/actions/${id}`, { approval: true });
  }

  revokeAction(id: string): Observable<any> {
    return this.httpClient.post(`${environment.cicCacheUrl}/actions/${id}`, { approval: false });
  }

  getHistoryByUser(id: string): Observable<any> {
    return this.httpClient.get(`${environment.cicCacheUrl}/history/${id}`);
  }

  getAccountDetailsFromMeta(userKey: string): Observable<any> {
    return this.httpClient.get(`${environment.cicMetaUrl}/${userKey}`, { headers: this.headers });
  }

  getUser(userKey: string): any {
    return this.httpClient.get(`${environment.cicMetaUrl}/${userKey}`, { headers: this.headers })
      .pipe(first()).subscribe(async res => {
      return Envelope.fromJSON(JSON.stringify(res)).unwrap();
    });
  }

  wrap(syncable: Syncable, signer: Signer): Promise<Envelope> {
    return new Promise<Envelope>(async (whohoo, doh) => {
      syncable.setSigner(signer);
      syncable.onwrap = async (env) => {
        if (env === undefined) {
          doh();
          return;
        }
        whohoo(env);
      };
      await syncable.sign();
    });
  }

  async loadAccounts(limit: number = 100, offset: number = 0): Promise<void> {
    this.resetAccountsList();
    const accountIndexAddress = await this.registry.getContractAddressByName('AccountRegistry');
    const accountIndexQuery = new AccountIndex(accountIndexAddress);
    const accountAddresses = await accountIndexQuery.last(await accountIndexQuery.totalAccounts());
    this.loggingService.sendInfoLevelMessage(accountAddresses);
    for (const accountAddress of accountAddresses.slice(offset, offset + limit)) {
      await this.getAccountByAddress(accountAddress, limit);
    }
  }

  async getAccountByAddress(accountAddress: string, limit: number = 100): Promise<Observable<AccountDetails>> {
    let accountSubject = new Subject<any>();
    this.getAccountDetailsFromMeta(await User.toKey(accountAddress)).pipe(first()).subscribe(async res => {
      const account = Envelope.fromJSON(JSON.stringify(res)).unwrap();
      this.accountsMeta.push(account);
      const accountInfo = account.m.data;
      await personValidation(accountInfo);
      accountInfo.balance = await this.tokenService.getTokenBalance(accountInfo.identities.evm['bloxberg:8996'][0]);
      accountInfo.vcard = vCard.parse(atob(accountInfo.vcard));
      this.accounts.unshift(accountInfo);
      if (this.accounts.length > limit) {
        this.accounts.length = limit;
      }
      this.accountsList.next(this.accounts);
      accountSubject.next(accountInfo);
    });
    return accountSubject.asObservable();
  }

  async getAccountByPhone(phoneNumber: string, limit: number = 100): Promise<Observable<AccountDetails>> {
    let accountSubject = new Subject<any>();
    this.getAccountDetailsFromMeta(await Phone.toKey(phoneNumber)).pipe(first()).subscribe(async res => {
      const response = Envelope.fromJSON(JSON.stringify(res)).unwrap();
      const address = '0x' + response.m.data;
      const account = await this.getAccountByAddress(address, limit);
      account.subscribe(result => {
        accountSubject.next(result);
      });
    });
    return accountSubject.asObservable();
  }

  resetAccountsList(): void {
    this.accounts = [];
    this.accountsList.next(this.accounts);
  }

  searchAccountByName(name: string): any { return; }
}
